==> fred_dnscheck-0.3.1.tar.gz: fred_dnscheck-0.3.1/dnscheck/process/nsset_checks.py <==
"""Process nsset check function."""

import datetime as dt
import logging
from collections.abc import Iterator
from typing import cast
from urllib.parse import urljoin

from fred_types import NssetId
from hermes import Email, EmailEnvelope, Reference
from regal.exceptions import NssetDoesNotExist
from sqlalchemy import exists, func, or_, select
from sqlalchemy.exc import IntegrityError

from dnscheck.constants import (
    CHECK_LEVEL_MAP,
    DomainCheckStatus,
    EmailType,
    MessageTemplate,
    NssetCheckStatus,
    ReferenceType,
)
from dnscheck.db import DomainCheck, Nsset, NssetCheck, atomic
from dnscheck.settings import APP
from dnscheck.utils import async_to_sync

_LOGGER = logging.getLogger(__name__)


def list_candidates() -> Iterator[NssetId]:
    """Get nsset check candidates.

    1. nssets that hasn't been checked yet
    2. nssets that hasn't been checked for the longest
       and that were last checked before min_check_interval

    Args:
        limit: Maximum number of returned nssets.
    """
    with atomic(expire_on_commit=False) as db:
        check_datetime_limit = dt.datetime.now(tz=dt.UTC) - APP.settings.min_check_interval
        last_check = (
            select(
                NssetCheck.nsset_id,
                func.max(NssetCheck.started_at).label("started_at"),
            )
            .group_by(NssetCheck.nsset_id)
            .subquery()
        )
        stmt = (
            select(Nsset.nsset_id)
            .outerjoin(last_check, Nsset.nsset_id == last_check.c.nsset_id)
            .outerjoin(
                NssetCheck,
                (Nsset.nsset_id == NssetCheck.nsset_id) & (NssetCheck.started_at == last_check.c.started_at),
            )
            .where(
                NssetCheck.started_at.is_(None) | NssetCheck.finished_at.is_not(None),  # there is no running check
                last_check.c.started_at.is_(None) | (NssetCheck.finished_at < check_datetime_limit),
            )
            .order_by(last_check.c.started_at.asc().nulls_first())
        )
        for nsset_id in db.execute(stmt).scalars():
            yield NssetId(nsset_id)


def start() -> None:
    """Start nsset checks until concurrency limit is reached."""
    with atomic() as sess:
        domain_checks = sess.execute(
            select(func.count(DomainCheck.id)).where(
                DomainCheck.status.in_({DomainCheckStatus.CREATED, DomainCheckStatus.STARTED})
            )
        ).scalar_one()

    candidates = list_candidates()
    while domain_checks < APP.settings.max_domain_checks:
        try:
            nsset_id = next(candidates)
            domain_checks += start_one(nsset_id)
        except StopIteration:
            break


@async_to_sync
async def start_one(nsset_id: NssetId) -> int:
    """Start nsset check.

    Args:
        nsset_id: Nsset id to be checked.

    Returns:
        Number of created domain checks.
    """
    try:
        info = await APP.nsset_client.get_nsset_info(nsset_id)
    except NssetDoesNotExist:
        # Nsset has been deleted since last synchronization
        _LOGGER.debug("Nsset %s does not exist anymore. Aborting nsset check.", nsset_id)
        return 0

    try:
        with atomic() as sess:
            nsset_check = NssetCheck(
                nsset_id=nsset_id,
                nsset_history_id=info.nsset_history_id,
                handle=info.nsset_handle,
                nameservers=[dns_host.model_dump(mode="json") for dns_host in info.dns_hosts],
            )
            sess.add(nsset_check)
            domain_checks = []
            async for domain in APP.domain_client.list_by_nsset(nsset_id):
                domain_checks.append(DomainCheck(domain_id=domain.id, domain=domain.fqdn, nsset_check=nsset_check))
            if domain_checks:
                sess.add_all(domain_checks)
                _LOGGER.info(
                    "Started nsset check of %s with %s linked domains", nsset_check.nsset_id, len(domain_checks)
                )
            else:
                # there is nothing to check, we finish the check immediately
                nsset_check.finished_at = func.now()
                nsset_check.status = NssetCheckStatus.FINISHED
                _LOGGER.info("Skipped nsset check of %s with no linked domains", nsset_check.nsset_id)

            return len(domain_checks)
    except IntegrityError:
        # nsset check has already been started (probably by another dnscheck instance)
        return 0


def finish() -> None:
    """Finish nsset checks."""
    while finish_one():
        pass


@async_to_sync
async def finish_one() -> bool:
    """Finish a single nsset check.

    This includes sending email notifications to technical contacts.
    """
    with atomic() as db:
        stmt = (
            select(NssetCheck)
            .where(
                NssetCheck.status == NssetCheckStatus.STARTED,
                ~exists(DomainCheck.id).where(
                    DomainCheck.nsset_check_id == NssetCheck.id,
                    DomainCheck.status.in_((DomainCheckStatus.CREATED, DomainCheckStatus.STARTED)),
                ),
            )
            .with_for_update(skip_locked=True)
            .limit(1)
        )
        nsset_check = db.execute(stmt).scalar_one_or_none()

        if nsset_check is None:
            return False

        nsset_check.status = NssetCheckStatus.FINISHED
        nsset_check.finished_at = dt.datetime.now(tz=dt.UTC)

        # Maximum allowed report level is 4. Any bigger level is considered to be 4.
        nsset = await APP.nsset_client.get_nsset_info(NssetId(nsset_check.nsset_id))
        check_level = min(cast(int, nsset.technical_check_level), 4)
        result_levels = CHECK_LEVEL_MAP[check_level]

        if not result_levels:
            # There is nothing to check, finish right away
            return True

        # Select total number of messages for each test result level
        stmt_total = (
            select(*[func.sum(DomainCheck.summary[level].as_integer()).label(level) for level in result_levels])
            .where(DomainCheck.nsset_check_id == nsset_check.id)
            .group_by(DomainCheck.nsset_check_id)
        )
        results_total = db.execute(stmt_total).one()

        if sum([r for r in results_total if r is not None]) > 0:
            # Select domain checks with the worst results
            stmt_worst = (
                select(DomainCheck)
                .where(
                    DomainCheck.nsset_check_id == nsset_check.id,
                    or_(False, *[DomainCheck.summary[level].as_integer() > 0 for level in result_levels]),
                )
                .order_by(
                    *[DomainCheck.summary[level].as_integer().desc().nulls_last() for level in result_levels],
                )
                .limit(APP.settings.max_notified_domains)
            )
            worst_domain_checks = db.execute(stmt_worst).scalars().all()

            context = {
                "nsset_handle": nsset.nsset_handle,
                "nsset_registrar": nsset.sponsoring_registrar,
                "started_at": nsset_check.started_at,
                "finished_at": nsset_check.finished_at,
                "total": {level.value: getattr(results_total, level) or 0 for level in result_levels},
                "bad_domains": [
                    {
                        "fqdn": check.domain,
                        "url": urljoin(APP.settings.zonemaster_api_url, f"/en/result/{check.test_id}"),
                    }
                    for check in worst_domain_checks
                ],
            }

            if APP.email_client:
                envelopes = []
                for contact_ref in nsset.technical_contacts:
                    contact = await APP.contact_client.get_contact_info(contact_ref.id)
                    envelopes.append(
                        EmailEnvelope(
                            message=Email(
                                recipient=contact.notify_emails or contact.emails,  # type: ignore[arg-type]
                                subject_template=MessageTemplate.DNSCHECK_SUBJECT,
                                body_template=MessageTemplate.DNSCHECK_BODY,
                                type=EmailType.DNSCHECK,
                                context=context,
                            ),
                            references=[
                                Reference(type=ReferenceType.CONTACT, value=contact_ref.id),
                                Reference(type=ReferenceType.NSSET, value=nsset.nsset_id),
                            ],
                        )
                    )
                _LOGGER.info("Sending email notifications about nsset %s.", nsset.nsset_id)
                async for message_id in APP.email_client.batch_send(envelopes):
                    _LOGGER.info("Message %s was sent", message_id)
            else:
                _LOGGER.info("Would send email notifications about nsset %s, but emails are disabled.", nsset.nsset_id)

        return True

==> fred_dnscheck-0.4.2.tar.gz: fred_dnscheck-0.4.2/dnscheck/process/nsset_checks.py <==
"""Process nsset check function."""

import datetime as dt
import logging
from collections.abc import Iterator
from typing import cast
from urllib.parse import urljoin

from fred_types import NssetId
from hermes import Email, Reference
from hermes.client import SendFailed
from regal.exceptions import ContactDoesNotExist, NssetDoesNotExist
from sqlalchemy import exists, func, or_, select
from sqlalchemy.exc import IntegrityError

from dnscheck.constants import (
    CHECK_LEVEL_MAP,
    DomainCheckStatus,
    EmailType,
    MessageTemplate,
    NssetCheckStatus,
    ReferenceType,
)
from dnscheck.db import DomainCheck, Nsset, NssetCheck, atomic
from dnscheck.settings import APP
from dnscheck.utils import defer_errors

_LOGGER = logging.getLogger(__name__)


def list_candidates() -> Iterator[NssetId]:
    """Get nsset check candidates.

    1. nssets that hasn't been checked yet
    2. nssets that hasn't been checked for the longest
       and that were last checked before min_check_interval

    Args:
        limit: Maximum number of returned nssets.
    """
    with atomic(expire_on_commit=False) as db:
        check_datetime_limit = dt.datetime.now(tz=dt.UTC) - APP.settings.min_check_interval
        last_check = (
            select(
                NssetCheck.nsset_id,
                func.max(NssetCheck.started_at).label("started_at"),
            )
            .group_by(NssetCheck.nsset_id)
            .subquery()
        )
        stmt = (
            select(Nsset.nsset_id)
            .outerjoin(last_check, Nsset.nsset_id == last_check.c.nsset_id)
            .outerjoin(
                NssetCheck,
                (Nsset.nsset_id == NssetCheck.nsset_id) & (NssetCheck.started_at == last_check.c.started_at),
            )
            .where(
                NssetCheck.started_at.is_(None) | NssetCheck.finished_at.is_not(None),  # there is no running check
                last_check.c.started_at.is_(None) | (NssetCheck.finished_at < check_datetime_limit),
            )
            .order_by(last_check.c.started_at.asc().nulls_first())
        )
        for nsset_id in db.execute(stmt).scalars():
            yield NssetId(nsset_id)


async def start() -> None:
    """Start nsset checks until concurrency limit is reached."""
    with atomic() as sess:
        domain_checks = sess.execute(
            select(func.count(DomainCheck.id)).where(
                DomainCheck.status.in_({DomainCheckStatus.CREATED, DomainCheckStatus.STARTED})
            )
        ).scalar_one()

    candidates = list_candidates()
    while domain_checks < APP.settings.max_domain_checks:
        try:
            nsset_id = next(candidates)
            domain_checks += await start_one(nsset_id)
        except StopIteration:
            break


async def start_one(nsset_id: NssetId) -> int:
    """Start nsset check.

    Args:
        nsset_id: Nsset id to be checked.

    Returns:
        Number of created domain checks.
    """
    try:
        info = await APP.nsset_client.get_nsset_info(nsset_id)
    except NssetDoesNotExist:
        # Nsset has been deleted since last synchronization
        _LOGGER.debug("Nsset %s does not exist anymore. Aborting nsset check.", nsset_id)
        return 0

    try:
        with atomic() as sess:
            nsset_check = NssetCheck(
                nsset_id=nsset_id,
                nsset_history_id=info.nsset_history_id,
                handle=info.nsset_handle,
                nameservers=[dns_host.model_dump(mode="json") for dns_host in info.dns_hosts],
            )
            sess.add(nsset_check)
            domain_checks = []
            async for domain in APP.domain_client.list_by_nsset(nsset_id):
                if APP.settings.domain_zones:
                    # Check whether FQDN has ".zone" suffix
                    if domain.fqdn.endswith(tuple(f".{zone}" for zone in APP.settings.domain_zones)):
                        domain_checks.append(
                            DomainCheck(domain_id=domain.id, domain=domain.fqdn, nsset_check=nsset_check)
                        )
                else:
                    domain_checks.append(DomainCheck(domain_id=domain.id, domain=domain.fqdn, nsset_check=nsset_check))
            if domain_checks:
                sess.add_all(domain_checks)
                _LOGGER.info(
                    "Started nsset check of %s with %s linked domains", nsset_check.nsset_id, len(domain_checks)
                )
            else:
                # there is nothing to check, we finish the check immediately
                nsset_check.finished_at = func.now()
                nsset_check.status = NssetCheckStatus.FINISHED
                _LOGGER.info("Skipped nsset check of %s with no linked domains", nsset_check.nsset_id)

            return len(domain_checks)
    except IntegrityError:
        # nsset check has already been started (probably by another dnscheck instance)
        return 0


async def finish() -> None:
    """Finish nsset checks."""
    with atomic() as db:
        stmt = select(NssetCheck.id).where(
            NssetCheck.status == NssetCheckStatus.STARTED,
            ~exists(DomainCheck.id).where(
                DomainCheck.nsset_check_id == NssetCheck.id,
                DomainCheck.status.in_((DomainCheckStatus.CREATED, DomainCheckStatus.STARTED)),
            ),
        )
        nsset_checks = db.execute(stmt).scalars().all()

    for check_id in nsset_checks:
        await defer_errors(finish_one)(check_id)


async def finish_one(nsset_check_id: int) -> None:
    """Finish a single nsset check.

    This includes sending email notifications to technical contacts.
    """
    with atomic() as db:
        stmt = (
            select(NssetCheck)
            .where(
                NssetCheck.id == nsset_check_id,
                NssetCheck.status == NssetCheckStatus.STARTED,
                ~exists(DomainCheck.id).where(
                    DomainCheck.nsset_check_id == NssetCheck.id,
                    DomainCheck.status.in_((DomainCheckStatus.CREATED, DomainCheckStatus.STARTED)),
                ),
            )
            .with_for_update(skip_locked=True)
            .limit(1)
        )
        nsset_check = db.execute(stmt).scalar_one_or_none()

        if nsset_check is None:
            return

        nsset_check.status = NssetCheckStatus.FINISHED
        nsset_check.finished_at = dt.datetime.now(tz=dt.UTC)

        try:
            # Maximum allowed report level is 4. Any bigger level is considered to be 4.
            nsset = await APP.nsset_client.get_nsset_info(NssetId(nsset_check.nsset_id))
        except NssetDoesNotExist:
            # Nsset has been deleted since the check start
            _LOGGER.warning("Nsset %s does not exist anymore. Aborting nsset check.", nsset_check.nsset_id)
            db.delete(nsset_check)
            return

        check_level = min(cast(int, nsset.technical_check_level), 4)
        result_levels = CHECK_LEVEL_MAP[check_level]

        if not result_levels:
            # There is nothing to check, finish right away
            return

        # Select total number of messages for each test result level
        stmt_total = (
            select(*[func.sum(DomainCheck.summary[level].as_integer()).label(level) for level in result_levels])
            .where(DomainCheck.nsset_check_id == nsset_check.id)
            .group_by(DomainCheck.nsset_check_id)
        )
        results_total = db.execute(stmt_total).one_or_none()

        if results_total and sum([r for r in results_total if r is not None]) > 0:
            # Select domain checks with the worst results
            stmt_worst = (
                select(DomainCheck)
                .where(
                    DomainCheck.nsset_check_id == nsset_check.id,
                    or_(False, *[DomainCheck.summary[level].as_integer() > 0 for level in result_levels]),
                )
                .order_by(
                    *[DomainCheck.summary[level].as_integer().desc().nulls_last() for level in result_levels],
                )
                .limit(APP.settings.max_notified_domains)
            )
            worst_domain_checks = db.execute(stmt_worst).scalars().all()

            context = {
                "nsset_handle": nsset.nsset_handle,
                "nsset_registrar": nsset.sponsoring_registrar,
                "started_at": nsset_check.started_at.isoformat(),
                "finished_at": nsset_check.finished_at.isoformat(),
                "total": {level.value: getattr(results_total, level) or 0 for level in result_levels},
                "bad_domains": [
                    {
                        "fqdn": check.domain,
                        "url": urljoin(
                            APP.settings.zonemaster_api_url,
                            f"/{APP.settings.zonemaster_language}/result/{check.test_id}",
                        ),
                    }
                    for check in worst_domain_checks
                ],
            }

            if APP.email_client:
                _LOGGER.info("Sending email notifications about nsset %s.", nsset.nsset_id)
                for contact_ref in nsset.technical_contacts:
                    try:
                        contact = await APP.contact_client.get_contact_info(contact_ref.id)
                    except ContactDoesNotExist:
                        _LOGGER.warning(
                            "Contact %s does not exist. Skipping.",
                            contact_ref.id,
                        )
                        continue

                    for email in contact.notify_emails or contact.emails:
                        message = Email(
                            recipient=email,
                            subject_template=MessageTemplate.DNSCHECK_SUBJECT,
                            body_template=MessageTemplate.DNSCHECK_BODY,
                            type=EmailType.DNSCHECK,
                            context=context,
                        )
                        try:
                            message_id = await APP.email_client.send(
                                message,
                                references=[
                                    Reference(type=ReferenceType.CONTACT, value=contact_ref.id),
                                    Reference(type=ReferenceType.NSSET, value=nsset.nsset_id),
                                ],
                            )
                            _LOGGER.info(
                                "Message %s to %s (contact %s) about nsset %s was sent.",
                                message_id,
                                email,
                                contact.contact_handle,
                                nsset.nsset_id,
                            )
                        except SendFailed:
                            _LOGGER.warning(
                                "Message to %s (contact %s) about nsset %s failed. Skipping.",
                                email,
                                contact.contact_handle,
                                nsset.nsset_id,
                            )
            else:
                _LOGGER.info("Would send email notifications about nsset %s, but emails are disabled.", nsset.nsset_id)
